;;;; This file is part of XMLTools
;;;; Copyright (C) 2004 Sergey Poznyakoff
;;;;
;;;; XMLTools is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; XMLTools is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with XMLTools; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;

(define-module (xmltrans lingua)
  #:export (begin-processing parse-command-line end-processing))

(use-modules (xmltools xmltrans)
	     (xmltools gamlet)
	     (xmltools dict)
	     (ice-9 format)
             (ice-9 rdelim)
	     (ice-9 getopt-long))
(use-syntax (ice-9 syncase))

;;;; Command line options
(define base-href-option #f)
(define prefix-option #f)

(define config-file-name "lingua.conf")
(define config-must-exist #f)

(define output-dir-option #f)
(define tmp-dir-option #f)

;;;; Global variables

;;; Keywords
(define keywords "")

;;; Index of the current item within the item group (e.g. ROW, ENUMERATE, etc)
(define item-index 0)
;;; Index of the current row within the row group (TABULAR)
(define row-index 0)
;;; Maximum number of items per row in the current table
(define max-items 0)
;;; Row index of the last separator row in the current table
(define separator-row 0)
;;; Keeps the index of the row ending the row span
(define rowspan-end 0)


;;; Dictionary of file name translations
(define reference-translation-list '())

(define-macro (define-reference-translation file-name reference)
  `(set! reference-translation-list (cons
				     (cons ,file-name ,reference)
				     reference-translation-list)))
(define (file-name->reference name)
  (let ((elem (assoc name reference-translation-list)))
    (if elem
	(cdr elem)
	name)))


;;; Dictionary of cross-reference identifiers
;;; Each element is a cons:
;;;  (ID . #(FILENAME (cons LOCUS-FILE LOCUS-LINE)))
(define xref-id-list '())

;;; Retrieve a cross-reference object by its identifier
(define-macro (xref-retrieve id)
  `(assoc ,id xref-id-list))

;;; Create a cross-reference object
(define-syntax make-xref
  (syntax-rules ()
    ((make-xref id) (make-xref id
			   (current-file-name)
			   (xmltrans:get-input-location)))
    ((make-xref id file-name) (if (not file-name)
				(cons id #f)
				(make-xref id file-name
					 (xmltrans:get-input-location))))
    ((make-xref id file-name loc) (cons id (vector file-name loc)))))

(define-macro (xref-name id)
  `(car ,id))

;;; Return #t if xref ID is defined
(define-macro (xref-defined? id)
  `(cdr ,id))

;;; Return the target file name associated with the xref ID
(define-macro (xref-file-name id)
  `(vector-ref (cdr ,id) 0))

;;; Return the location of the definition of xref ID
(define-macro (xref-source-location id)
  `(vector-ref (cdr ,id) 1))

;;; Define an already registered cross-reference by associating with it
;;; the current file name and input location.
;;;
(define-macro (xref-define id)
  `(set-cdr! ,id (vector (current-file-name) (xmltrans:get-input-location))))

;;; Register a cross reference. Optional argument gives the target file name
;;; (or #f to create undefined (forward) cross-reference.
(define (xref-register id . rest)
  (cond
   ((assoc id xref-id-list) =>
    (lambda (x)
      (xmltrans:parse-error #f "Reference ID " (xref-name x) " already defined")
      (xmltrans:parse-error (xref-source-location x)
			    "This is the location of previos definition")))
   (else
    (set! xref-id-list
	  (cons
	   (cond
	    ((null? rest)
	     (make-xref id))
	    ((= (length rest) 1)
	     (make-xref id (car rest)))
	    (else
	     (throw 'misc-error 'menu-append
		    "Invalid number of optional arguments"
		    (list (xmltrans:input-file-name)
			  (xmltrans:input-line-number)) #f)))
	   xref-id-list)))))

;;; Resolve forward cross reference. This function is called from gamlet
;;; postprocess handler.
(define (resolve-forward-xref ref-name location)
  (let ((id (xref-retrieve ref-name)))
    (cond
     ((and id (xref-defined? id))
      (display (reference-name (file-name->reference (xref-file-name id))))
      (display "#")
      (display ref-name))
     (else
      (xmltrans:parse-error location "Undefined reference " ref-name)))))
      

;;;; Temporary gettext
(define gettext-trans '())

(define-public (lingua:gettext text)
  (cond
   ((assoc text gettext-trans) =>
    (lambda (x)
      (cdr x)))
   (else
    text)))

(xmltrans:end-tag
 "MSGSTR"
 (tag attrs text)
 (let ((id (xmltrans:attr #f "ID")))
   (if id
       (set! gettext-trans (cons
			    (cons id text)
			    gettext-trans))
       (xmltrans:parse-error #f "MSGSTR tag missing ID")))
 #f)


;;;; Menu support
(define (menu-append tag object label)
  (let ((menu (xmltrans:attr tag "MENU")))
    (cond
     ((not menu)
      (throw 'misc-error 'menu-append
	     "~A:~A: Parent element does not have MENU element"
	     (list (xmltrans:input-file-name)
		   (xmltrans:input-line-number)) #f)))
    (xmltrans:set-attr
     tag "MENU"
     (append menu
	     (list (cons object label))))))
		       

(define (find-file-in-menu filename . rest)
  (let ((menu (if (null? rest) (lingua:attr "LINGUA" "MENU") (car rest))))
    (call-with-current-continuation
     (lambda (return)
       (for-each
	(lambda (elem)
	  (let ((object (car elem)))
	    (cond
	     ((string? object)
	      (if (string=? object filename)
		  (return (cdr elem))))
	     ((list? object)
	      (let ((x (find-file-in-menu filename object)))
		(if x
		    (return x)))))))
	menu)
       #f))))

(define (print-menu-map menu)
  (display "<UL>\n")
  (for-each
   (lambda (elem)
     (let ((object (car elem))
	   (label (cdr elem)))
       (cond
	((string? object)
	 (format #t "<LI><A HREF=\"~A\">~A</A></LI>"
		 (reference-name object)
		 label))
	((list? object)
	 (cond
	  ((string-null? label)
	   (format #t "<LI><A HREF=\"~A\">~A</A>\n"
		   (reference-name (caar object)) (cdar object))
	   (print-menu-map (cdr object)))
	  (else
	   (display "<LI>")
	   (display label)
	   (print-menu-map object)
	   (display "</LI>")))))))
   menu)
  (display "</UL>"))

(define (current-menu-item? item)
  (and
   (string? (car item))
   (string=? (car item) (gamlet:current-output-file-name))))

(define (put-menu-item item level)
  (cond
   ((procedure? (car item))
    ((car item)))
   ((list? (car item))
    (if (and (string? (cdr item)) (not (string-null? (cdr item))))
	(format #t
		"<TR><TD class=\"group\"><DIV ALIGN=\"CENTER\">~A</DIV></TD></TR>\n"
		(cdr item)))
    (cond
     ((= 0 level)
      (out-menu (car item) (1+ level)))
     (else
      (put-menu-item (caar item) level)
      (if (has-current-menu-item? (car item))
	  (out-menu (cdr (car item)) (1+ level))))))
   ((string? (car item))
    (display "<TR><TD>")
    (if (> level 0)
	(do ((i 1 (1+ i)))
	    ((= i level))
	  (display "&gt;")))
    (cond
     ((current-menu-item? item)
      (display (cdr item)))
     (else
      (format #t "<A HREF=\"~A\">~A</A>"
	      (reference-name (car item))
	      (cdr item))))
    (display "</TD></TR>\n"))
   (else
    (throw 'misc-error 'put-menu-item
	   "Unrecognized menu item ~S" (list item) #f))))
  
(define (has-current-menu-item? menu)
  (call-with-current-continuation
   (lambda (exit)
     (for-each
      (lambda (item)
	(cond
	 ((list? (car item))
	  (if (has-current-menu-item? (car item))
	      (exit #t)))
	 (else
	  (if (current-menu-item? item)
	      (exit #t)))))
      menu)
     #f)))
     

(define (out-menu menu level)
;  (display "MENU " (current-error-port))
;  (write menu (current-error-port))
;  (display "\n" (current-error-port))

  (for-each
   (lambda (item)
     (put-menu-item item level))
   menu))
	      
		


;;;; Utility functions

(define-macro (defn-put func-name arg-name)
  `(lambda ( . rest )
		(cond
		 ((not (null? rest))
		  (set! ,arg-name (cons (car rest) ,arg-name))
		  (apply ,func-name (cdr rest))))))

(define (left-pad-string str n chr)
  (let ((len (string-length str)))
    (cond
     ((< len n)
      (string-append (make-string (- n len) chr) str))
     (else
      str))))

(define (create-file-name prefix)
  (let ((lang (let ((l (xmltrans:attr "LINGUA" "LANG")))
		(if l
		    (string-append l "/")
		    "")))
	(suffix ".html"))
    (cond
     ((get-attribute-value #f "PAGENO" 1+) =>
      (lambda (n)
	(if (= 0 n)
	    (string-append lang prefix suffix)
	    (string-append lang prefix "_"
			   (left-pad-string (number->string n) 2 #\0)
			   suffix))))
     (else
      (string-append lang prefix suffix)))))

;;;; Look up for attribute NAME in the current tag stack starting
;;;; from TAG or from the stack top if TAG is #f. Return attribute
;;;; value if found or #f otherwise
(define (get-attribute-value tag name . rest)
  (let ((proc (and (not (null? rest)) (car rest))))
    (call-with-current-continuation
     (lambda (exit)
       (letrec ((getval
		 (lambda (name attrs)
		   (let ((item (assoc name attrs)))
		     (if item
			 (let ((val (cdr item)))
			   (if proc
			       (set-cdr! item (proc val)))
			   (exit val)))))))
	 (xmltrans:for-each-tag
	  (lambda (tag-name tag-attrs)
	    (if tag
		(cond
		 ((string=? tag-name tag)
		  (set! tag #f)
		  (getval name tag-attrs)))
		;; else:
		(getval name tag-attrs)))))
       #f))))

(define (reference-name name)
  (cond
   ((char=? (string-ref name 0) #\/)
    (substring name 1))
   ((or (and (xmltrans:tag-exists? "LINGUA")
	     (xmltrans:attr "LINGUA" "TARGETDIR"))
	(and (lingua:tag-exists? "LINGUA")
	     (lingua:attr "LINGUA" "TARGETDIR"))
	(and (lingua:tag-exists? "DIRECTORY")
	     (lingua:attr "DIRECTORY" "TARGET"))) =>
    (lambda (pref)
      (string-append pref "/" name)))
   (prefix-option
    (string-append prefix-option "/" name))
   (else
    name)))

(define (current-file-name)
  (xmltrans:attr "PAGE" "FILE-NAME"))

;;;; Footnote support
(define (add-footnote obj text id)
  (cond
   ((and id (assoc id (or (xmltrans:attr obj "FOOTNOTE-IDS") '()))) =>
    (lambda (n)
      (cdr n)))
   (else
    (let ((fl (or (xmltrans:attr obj "FOOTNOTES") '())))
      (xmltrans:set-attr obj "FOOTNOTES" (append fl (list text)))
      (let ((fn (1+ (length fl))))
	(if id
	    (xmltrans:set-attr obj
			       "FOOTNOTE-IDS"
			       (cons
				(cons id fn)
				(or (xmltrans:attr obj "FOOTNOTE-IDS") '()))))
	fn)))))

(define (footnote-name n)
  (string-append "FNNAME:" (number->string n)))

(define (footnote-href n)
  (string-append "FNHREF:" (number->string n)))


(define-macro (map-footnotes fl proc)
  `(cond
    (,fl
     (,proc "<div class=\"footnotes\">")
     (do ((cur ,fl (cdr cur))
	  (n 1 (1+ n)))
	 ((null? cur))
       (,proc (format #f "<p><a name=\"~A\" href=\"~A#~A\"><sup>~A)</sup></a> "
		      (footnote-href n)
		      (reference-name (file-name->reference (current-file-name)))
		      (footnote-name n)
		      n))
       (,proc (car cur))
       (,proc "</p>\n"))
     (,proc "</div>\n"))))

(define (print-footnotes fl)
  (map-footnotes fl display))
	   
;;;; Create a template for a page
(define (page-template title) 
  (string-concat
   "<BODY>
<TABLE class=\"main\" WIDTH=\"99%\" ALIGN=\"CENTER\" CELLSPACING=\"0\" CELLPADDING=\"10%\" RULES=\"COLS\">
<TR>
 <TD VALIGN=top CLASS=\"contentpanel\">
  <H1 ALIGN=CENTER>"
   title
   "</H1>
  <HR>"))

(define (get-cgi-bin)
  (or
   (and (lingua:tag-exists? "DIRECTORY")
	(lingua:attr "DIRECTORY" "CGIBIN"))
   "/cgi-bin"))
    
(define (get-email)
  (cond
   ((lingua:attr "AUTHOR" "EMAIL") =>
    (lambda (x)
      x))
   (else
    "gray@gnu.org")))

(define (get-base-href)
  (cond
   ((xmltrans:attr "LINGUA" "BASE") =>
    (lambda (x)
      x))
   (else
    base-href-option)))

(define (open-page title)
  (format #t "\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">	
<HTML>
<HEAD>
 <TITLE>~A</TITLE>
 <LINK REV=\"made\" HREF=\"mailto:~A\">
 <META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; CHARSET=UTF-8\">
 <META NAME=\"GENERATOR\" CONTENT=\"~A\">
 <META NAME=\"keywords\" LANG=\"~A\" CONTENT=\"~A\">
 <META NAME=\"ROBOTS\" CONTENT=\"INDEX,NOFOLLOW\">
 <BASE HREF=\"~A\">
 <LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"~A\">
</HEAD>"
	  title
	  (get-email)
	  xmltrans:generator
	  (xmltrans:attr "LINGUA" "LANG")
	  keywords
	  (get-base-href)
	  (reference-name "mainstyle.css"))

  (format #t "
<BODY>
<TABLE class=\"main\" id=\"top\" WIDTH=\"99%\" ALIGN=\"CENTER\" CELLSPACING=\"0\" CELLPADDING=\"10%\" RULES=\"COLS\">
<TR>
 <TD VALIGN=top CLASS=\"contentpanel\">
  <H1 ALIGN=CENTER>~A</H1>
  <HR>"	  
	  (or (xmltrans:attr #f "HEADER") title)))
  
(define (close-page)
  (print-footnotes (xmltrans:attr "PAGE" "FOOTNOTES"))
  (format #t "
   {BOTTOM-NAVBAR}
 </TD>
 <TD class=\"navbar\" VALIGN=TOP width=\"20%\">
  <IMG SRC=\"~A\" ALIGN=\"top\" ALT=\"~A\"><P>
   <TABLE class=\"navbar\" WIDTH=\"100%\" ALIGN=\"LEFT\" CELLSPACING=\"0\" CELLPADDING=\"10%\" BORDER=\"0\">
   {TOC}

   <TR><TD class=\"group\"><DIV ALIGN=\"CENTER\">~A</DIV></TD></TR>
   <TR><TD><A HREF=\"~A\">~A</A></TD></TR>
   <TR><TD><A HREF=\"http://validator.w3.org/check/referer\">
    <img border=\"0\" src=\"graphics/valid-html401.png\"
          alt=\"[ Valid HTML 4.01! ]\" height=\"31\" width=\"88\"></A></TD></TR>

   <TR><TD><a href=\"http://jigsaw.w3.org/css-validator/check/referer\">
    <img border=\"0\" src=\"graphics/valid-css.png\" 
         alt=\"[ Valid CSS! ]\" height=\"31\" width=\"88\"></A></TD></TR>

   <TR><TD><a href=\"http://www.gnu.org\">
    <img border=\"0\" src=\"graphics/button-powered_by_gnu.png\"
         alt=\"[ Powered by GNU ]\" height=\"47\" width=\"102\"></a></TD></TR> 

   {INFO}
	  
  </TABLE>
 </TD>
</TR>
<TR>
 <TD colspan=\"2\">
 "
	  (lingua:attr "LOGO" "SRC")
	  (lingua:attr "LOGO" "ALT")
	  (lingua:gettext "Information")
	  (lingua:attr "AUTHOR" "HOMEPAGE")
	  (lingua:gettext "Home Page"))

  (display "<div class=\"bottom\">\n")
  (cond
   ((xmltrans:attr "PAGE" "BOTTOM") =>
    (lambda (x)
      (newline)
      (display x)
      (newline))))

  (cond
   ((lingua:attr "COPYRIGHT" "TEXT") =>
    (lambda (x)
      (display x)))
   ((lingua:attr "AUTHOR" "CONTACT") =>
    (format #t "<P>~A <A HREF=\"mailto:~A\">~A</A>;</P>"
	    (lingua:gettext "For information about this site contact")
	    (get-email)
	    (lingua:attr "AUTHOR" "CONTACT")))
   (else
    (xmltrans:parse-warning #f
			    "Neither AUTHOR.CONTACT nor COPYRIGHT are defined")))
      
  
  (display "</div>\n
 </TD>
</TR>
</TABLE>
</BODY>
</HTML>"))


(define (set-anchor attrs . rest)
  (cond
   ((xmltrans:attr attrs "ID") =>
    (lambda (xref-name)
      (let ((id (xref-retrieve xref-name)))
	(if id
	    (xref-define id)
	    (xref-register xref-name)))
      (format #f "<A NAME=\"~A\">~A</A>"
	      xref-name
	      (if (not (null? rest)) (car rest) ""))))
   (else
    #f)))



;;;; Handlers

;;;; Default handlers

(define tag-list '())

(define (lingua:default-start tag attr)
  (xmltrans:set-attr #f "__START__" 1)
  #f)

(xmltrans:set-default-start-handler lingua:default-start)

(define (lingua:default-end tag attr text)
  (if (xmltrans:attr attr "__START__")
      (xmltrans:parse-error #f "Unhandled element " tag))
  (set! tag-list
	(cons
	 (xmltrans:make-tag tag attr text)
	 tag-list))
  #f)

(xmltrans:set-default-end-handler lingua:default-end)

(define (lingua:handled-element-end tag attr text)
  (set! tag-list
	(cons
	 (xmltrans:make-tag tag attr text)
	 tag-list))
  #f)

(define-macro (lingua:tag-ignore name)
  `(xmltrans:set-end-handler (symbol->string ',name)
			     lingua:handled-element-end))


;; FIXME: duplicated from xmltrans
;;;; Returns #t if TAG exists on stack
(define-public (lingua:tag-exists? tag)
  (assoc tag tag-list))

(define-public (lingua:tag-attributes tag)
  (let ((item (if tag (assoc tag tag-list) (car tag-list))))
    (if (not item)
	(throw 'misc-error 'lingua:tag-attributes
	       "No such tag on stack ~S" (list tag) #f))
    (xmltrans:tag-ref #:attrs item)))

(define-public (lingua:attr tag name)
  (let ((attr (assoc name (lingua:tag-attributes tag))))
    (and attr (cdr attr))))

(define-public (lingua:set-attr tag name value)
  (let ((item (xmltrans:extract-item tag-list tag)))
    (if (not item)
	(throw 'misc-error 'lingua:set-attr
	       "No such tag on stack ~S" (list tag) #f))
    (let ((attr (assoc name (xmltrans:tag-ref #:attrs item))))
      (if attr
	  (set-cdr! attr value)
	  (xmltrans:tag-set! #:attrs
			     item
			     (cons
			      (cons name value)
			      (xmltrans:tag-ref #:attrs item)))))))


(define (replicate-menu from to)
  (cond
   ((xmltrans:attr from "MENU") =>
    (lambda (m)
      (xmltrans:set-attr to "MENU" m)))))


;;;; Element handlers

;;; Configuration file elements
(xmltrans:end-tag
 "DIRECTORY"
 (tag attr text)
 (for-each
  (lambda (a)
    (if (not (lingua:tag-exists? "DIRECTORY"))
	(set! tag-list (cons
			(xmltrans:make-tag "DIRECTORY")
			tag-list)))
    (lingua:set-attr "DIRECTORY" (car a) (cdr a)))
  attr)
 #f)

(xmltrans:end-tag
 "BASE"
 (tag attr text)
 (if (not base-href-option)
     (set! base-href-option (xmltrans:attr attr "HREF")))
 #f)

;;; Ignored tags
(lingua:tag-ignore LOGO)
(lingua:tag-ignore CONFIG)
(lingua:tag-ignore AUTHOR)
(lingua:tag-ignore LOCALIZATION)
 
;;; Simple tags
(xmltrans:replace-tag DFN dfn)
(xmltrans:replace-tag CMD code)
(xmltrans:replace-tag PROG b)
(xmltrans:replace-tag QUOTE i)
(xmltrans:replace-tag ITEMIZE dl)
(xmltrans:replace-tag PREF b)
(xmltrans:replace-tag SUFF b)
(xmltrans:replace-tag EMPH b)
(xmltrans:replace-tag SUBSCRIPT sub)
(xmltrans:replace-tag SUPERSCRIPT sup)
(xmltrans:replace-tag IT i)

(xmltrans:end-tag
 "NEWLINE"
 (tag attr text)
 (if (string-null? text)
     "\n"
     (list "\n" text "\n")))

;;;; INCLUDE
(xmltrans:end-tag
 "INCLUDE"
 (tag attr text)
 (let ((fname (xmltrans:attr attr "FILE")))
   (cond
    ((not fname)
     (xmltrans:parse-error #f "File name not specified"))
    ((string? fname)
     (replicate-menu #t #f)
     (xmltrans:parse-file fname)
     (replicate-menu #f #t))
    (else
     (xmltrans:parse-error #f "FILE must be a valid string"))))
 #f)

;;;; COPYRIGHT
(xmltrans:end-tag
 "COPYRIGHT"
 (tag attr text)
 (lingua:handled-element-end tag (cons (cons "TEXT" text) attr) text))

;;;; GUILE
(xmltrans:end-tag
 "GUILE"
 (tag attr text)
 (cond
  ((xmltrans:attr attr "SOURCE") =>
   (lambda (f)
     (load f))))
 (if (not (string-null? text))
     (eval-string text)))

;;;; LINGUA handlers

(xmltrans:start-tag
 "LINGUA"
 (tag attr)
  ;(set-locale (xmltrans:attr "LINGUA" "LANG"))
 (letrec ((dictref (lambda ()
		     (string-append
		      (get-cgi-bin)
		      "/dict.cgi?LANG="
		      (xmltrans:attr "LINGUA" "LANG")
		      "_"
		      (xmltrans:attr "LINGUA" "TER"))))
	  (navigation
	   (lambda ()
	     (let ((index (gamlet:output-file-index)))
	     (if (> index 0)
		 (put-menu-item (cons
				 (gamlet:output-file-name (- index 1))
				 (lingua:gettext "Previous Page"))
				0))
	     (if (< index (- (gamlet:output-file-count) 1))
		 (put-menu-item (cons
				 (gamlet:output-file-name (+ index 1))
				 (lingua:gettext "Next Page"))
				0))))))
			  
   (xmltrans:set-attr #f "DICT" dictref)
   (xmltrans:set-attr #f "NAVIGATION" navigation))
 
 (xmltrans:set-attr #f "MENU" '())
 "")

(xmltrans:end-tag
 "LINGUA"
 (tag attr text)
 (lingua:default-end tag attr text)
 #f)

;;; 

;;; PAGE

(define (page-title)
  (cond
   ((or (xmltrans:attr "PAGE" "TITLE")
	(xmltrans:attr "PAGE" "HEADER")) =>
	(lambda (x)
	  x))
   (else
    (xmltrans:parse-warning #f "Page title not defined")
    (string-append "UNDEFINED at "
		   (xmltrans:input-file-name)
		   ":"
		   (number->string
		    (xmltrans:input-line-number))))))

(define (page-menu tag)
  (menu-append
   tag
   (cond
    ((xmltrans:attr "PAGE" "REF") =>
     (lambda (x)
       (let ((fn (xmltrans:attr "LINGUA" x)))
	 (cond
	  ((not fn)
	   (xmltrans:parse-error #f "Function " x " not defined")
	   (current-file-name))
	  ((not (procedure? fn))
	   (xmltrans:parse-error #f x " is not a function")
	   (current-file-name))
	  (else
	   (let ((ref (fn)))
	     (define-reference-translation (current-file-name) ref)
	     ref))))))
    (else
     (current-file-name)))

   ;; label:
   (or (xmltrans:attr "PAGE" "HEADER")
       (xmltrans:attr "PAGE" "TITLE")
       (begin
	 (xmltrans:parse-warning #f "Page header not defined")
	 (string-append "UNDEFINED at "
			(xmltrans:input-file-name)
			":"
			(number->string
			 (xmltrans:input-line-number)))))))

(xmltrans:start-tag
 "PAGE"
 (tag attr)
 (xmltrans:set-attr
  #f
  "FILE-NAME"
  (create-file-name (or (get-attribute-value #f "PREFIX") "prefix")))
 (if (xmltrans:attr attr "HEADER")
     (page-menu #t)
     (let ((loc (xmltrans:get-input-location)))
       (letrec ((proc (lambda ()
			(xmltrans:parse-warning
			 loc
			 "PAGE element does not have HEADER attribute.")
			(xmltrans:set-attr "PAGE" "HEADER"
					   (xmltrans:attr "PAGE" "TITLE"))
			(page-menu (xmltrans:get-parent "PAGE")))))
	 (xmltrans:expect-next-tag "HEADER" proc))))
 (xmltrans:set-attr #f "MENU" '())
 #f)
	
(xmltrans:end-tag
 "PAGE"
 (tag attr text)
 (gamlet:with-output-to-file
  (current-file-name)
  (lambda ()
    (open-page (page-title))
    (display text)
    (close-page)))
 #f)

;;; KEYWORDS
(xmltrans:end-tag
 "KEYWORDS"
 (tag attr text)
 (set! keywords text)
 "")

;;; TITLE
(xmltrans:end-tag
 "TITLE"
 (tag attr text)
 (cond
  ((xmltrans:parent? "TABULAR")
   (string-append "<CAPTION>" text "</CAPTION>"))
  (else
   (xmltrans:parse-error #f "Misplaced TITLE")
   #f)))

;;; CHAPTER
(xmltrans:start-tag
 "CHAPTER"
 (tag attr)
 (xmltrans:set-attr #f "PAGENO" 0)
 (xmltrans:set-attr #f "MENU" '())
 #f) ;; Cannot have (set-anchor attr))

(xmltrans:end-tag
 "CHAPTER"
 (tag attr text)  
 (menu-append #t (xmltrans:attr attr "MENU") "")
 #f)

;;; GROUP
(xmltrans:start-tag
 "GROUP"
 (tag attr)
 (xmltrans:set-attr #f "MENU" '())
 #f)

(xmltrans:end-tag
 "GROUP"
 (tag attr text)
 (if (not (assoc "HIDDEN" attr))
     (menu-append #t
		  (xmltrans:attr attr "MENU")
		  (xmltrans:attr attr "TITLE")))
 #f)



(define (check-sectioning parent)
  (if (not (xmltrans:parent? parent))
      (xmltrans:parse-error #f "Incorrect sectioning. Expected "
			    parent
			    " but found "
			    (xmltrans:get-parent))))

;;; SECTION
(xmltrans:start-tag
 "SECTION"
 (tag attr)
 (check-sectioning "PAGE")
 (set-anchor attr))

(xmltrans:end-tag
 "SECTION"
 (tag attr text)
 (list "<div class=\"section\">" text "</div>"))

;;; SUBSECTION
(xmltrans:start-tag
 "SUBSECTION"
 (tag attr)
 (check-sectioning "SECTION")
 (set-anchor attr))

(xmltrans:end-tag
 "SUBSECTION"
 (tag attr text)
 (list "<div class=\"subsection\">" text "</div>"))

;;; SUBSUBSECTION
(xmltrans:start-tag
 "SUBSUBSECTION"
 (tag attr)
 (check-sectioning "SUBSECTION")
 (set-anchor attr))

(xmltrans:end-tag
 "SUBSUBSECTION"
 (tag attr text)
 (list "<div class=\"subsubsection\">" text "</div>"))

;;; ANCHOR
(xmltrans:end-tag
 "ANCHOR"
 (tag attr text)
 (set-anchor attr text))

;;; XREF
(xmltrans:start-tag
 "XREF"
 (tag attr)
 (cond
  ((assoc "URL" attr) =>
   (lambda (url)
     (format #f "<A HREF=\"~A\">" (cdr url))))
  ((assoc "REF" attr) =>
   (lambda (ref)
     (let* ((ref-name (cdr ref))
	    (id (xref-retrieve ref-name)))
       (cond
	((and id (xref-defined? id))
	 (string-append
	  "<A HREF=\""
	  (reference-name (file-name->reference (xref-file-name id)))
	  "#"
	  ref-name
	  "\">"))
	(else
	 (if (not id) 
	     (xref-register ref-name #f))
	 ;; The reference is registered but not defined
	 ;; Create a postprocess handler and insert its invocation to the
	 ;; output stream. The handler is defined within a closure that
	 ;; supplies target reference name and source location for
	 ;; resolve-forward-xref
	 (let ((location (xmltrans:get-input-location)))
	   (letrec ((name (string-append "@@" ref-name ":"))
		    (proc (lambda ()
			    (resolve-forward-xref ref-name location))))
	     (gamlet:set-handler name proc)
	     (let ((delim (gamlet:get-pp-delimiter)))
	       (format #f "<A HREF=\"~A~A~A\">"
		       (car delim) name (cdr delim))))))))))
  (else
   (xmltrans:parse-error #f "XREF element must have URL or REF attrbute")
   "<A HREF=\"NONEXISTENT\">")))

(xmltrans:end-tag
 "XREF"
 (tag attr text)
 (list text "</A>"))

;;; BOTTOM	     
(xmltrans:end-tag
 "BOTTOM"
 (tag attr text)
 (xmltrans:set-attr "PAGE" "BOTTOM" text)
 #f)

(define (footnote-parent)
  (if (xmltrans:tag-exists? "TABULAR") ;; FIXME
      "TABULAR"
      "PAGE"))

;;; FOOTNOTE
(xmltrans:end-tag
 "FOOTNOTE"
 (tag attr text)
 (cond
  ((not (xmltrans:tag-exists? "PAGE"))
   (xmltrans:parse-error #f
			 "Incorrect sectioning. FOOTNOTE may be used only inside PAGE"))
  (else
   (let ((n (add-footnote (footnote-parent) text (xmltrans:attr attr "ID"))))
     (format #f "<a name=\"~A\" href=\"~A#~A\"><sup>~A)</sup></a>"
	     (footnote-name n)
	     (reference-name (file-name->reference (current-file-name)))
	     (footnote-href n)
	     n)))))

;;; PARA
(xmltrans:start-tag
 "PARA"
 (tag attr)
 "\n<P>")

(xmltrans:end-tag
 "PARA"
 (tag attr text)
;;FIXME: This may generate empty <P> blocks
 (if (not (assoc "__CLOSED__" attr))
     (list text "</P>\n")
     text))

;;; HEADER

(define (format-header level text)
  (let ((pfx (set-anchor #f)))
    (format #f "\n~A<H~A ALIGN=\"CENTER\">~A</H~A>\n"
	    (or pfx "") level text level)))

(xmltrans:end-tag
 "HEADER"
 (tag attr text)
 (cond
  ((xmltrans:parent? "PAGE")
   (cond
    ((xmltrans:attr #t "HEADER")
     (xmltrans:parse-warning #f
			   "PAGE element already has HEADER defined")
     (xmltrans:set-attr "PAGE" "HEADER" text))
    (else
     (xmltrans:set-attr "PAGE" "HEADER" text)
     (page-menu (xmltrans:get-parent "PAGE"))))
    (set-anchor attr))
  ((xmltrans:parent? "SECTION")
   (format-header 2 text))
  ((xmltrans:parent? "SUBSECTION")
   (format-header 3 text))
  ((xmltrans:parent? "SUBSUBSECTION")
   (format-header 4 text))
  ((xmltrans:parent? "DEFLIST")
   (list "<DT>" text "</DT>"))
  (else
   (xmltrans:parse-error #f "Misplaced HEADER")
   #f)))

;;; 
;;; Split the given text string at split delimiters. Print each part
;;; except the first one in bold face and indent it one space from
;;; the previous one  
(define (item-split text)
  (cond
   ((xmltrans:attr "TABULAR" "SPLIT") =>
    (lambda (delim)
      (let ((l (string-split text (string-ref delim 0))))
	(if (null? (cdr l))
	    (car l)
	    (apply string-append
		   (cons
		    (car l)
		    (do ((cur (cdr l) (cdr cur))
			 (out '()))
			((null? cur)
			 out)
		      (set! out (append out (list "&nbsp;<B>" (car cur) "</B>"))))))))))
   (else
    text)))

;;; Print the contents of the current item using splitting rules.
;;;  SPLIT attribute introduces a split delimiter.
;;;  OR attribute introduces an "or" delimiter
(define (item-text text)
  (cond
   ((xmltrans:attr "TABULAR" "OR") =>
    (lambda (delim)
      (let ((l (string-split text (string-ref delim 0))))
	(if (null? (cdr l))
	    (item-split text)
	    (apply string-append
		   (cons
		    (item-split (car l))
		    (do ((cur (cdr l) (cdr cur))
			 (out '()))
			((null? cur)
			 out)
		      (set! out
			    (append
			     (list "&nbsp;" delim "&nbsp;"
				   (item-split (car cur))))))))))))
   (else
    (item-split text))))

;;; Remove any split delimiters from the given text string	   
(define (restore-single-word text)
  (cond
   ((xmltrans:attr "TABULAR" "SPLIT") =>
    (lambda (delim)
      (apply string-append (string-split text (string-ref delim 0)))))
   (else
    text)))

;;; Restore the "canonical" form of the item text by removing any split
;;; delimiters from it and gluing the resulting parts together. If the
;;; item has several parts separated by "or" delimiters, only the first
;;; part is acted upon.
(define (restore-word text)
  (cond
   ((xmltrans:attr "TABULAR" "OR") =>
    (lambda (delim)
      (restore-single-word (car (string-split text (string-ref delim 0))))))
   (else
    (restore-single-word text))))

;;; Create a dictionary reference for a given WORD. REFTEXT gives the
;;; reference text.
(define (dict-reference word reftext . rest)
  (let ((backref (and (not (null? rest)) (car rest))))
    (format #f
	    "<A ~AHREF=\"~A&amp;IDENT=~A\">~A</A>"
	    (if backref
		(string-append "NAME=\"id" (dict:encode-string word) "\" ")
		"")
	    ((xmltrans:attr "LINGUA" "DICT"))
	    (dict:encode-string word)
	    reftext)))

;;; Return true if we are on the very first data row in a table. Colon
;;; headings are not counted as data row.

(define (first-data-row?)
  (if (xmltrans:attr "TABULAR" "COLHEADING")
      (= row-index 1)
      (= row-index 0)))

;;; Return effective row index
;;; Rows are numbered from 1, column headings are not counted	   
(define (eff-row-index)
  (if (xmltrans:attr "TABULAR" "COLHEADING")
      row-index
      (1+ row-index)))

;;; Return effective column (item) index
;;; Columns are numbered from 1, row headings are not counted	   
(define (eff-col-index)
  (if (xmltrans:attr "TABULAR" "ROWHEADING")
      item-index
      (1+ item-index)))

;;; Table map structure: #( start-col end-col start-row end-row )
;;; FIXME: Originally it had 4th element: need-ref?
(define-macro (map-start-col map)
  `(vector-ref ,map 0))

(define-macro (map-end-col map)
  `(vector-ref ,map 1))

(define-macro (map-start-row map)
  `(vector-ref ,map 2))

(define-macro (map-end-row map)
  `(vector-ref ,map 3))

(define (parse-dictref str)
  (let ((map (vector -1 -1 -1 -1)))
    (letrec ((set-map (lambda (n x)
			(vector-set! map n (string->number (car x)))
			(vector-set! map (1+ n)
				     (if (or (null? (cdr x))
						    (string-null? (cadr x)))
					 0
					 (string->number (cadr x)))))))
      (for-each
       (lambda (s)
	 (let ((x (string-split s #\:)))
	   (cond
	    ((string=? (car x) "col")
	     (set-map 0 (cdr x)))
	    ((string=? (car x) "row")
	     (set-map 2 (cdr x)))
	    (else
	     (xmltrans:parse-warning #f "Unknown modifier " x)))))
       (string-split str #\;))
      (cond
       ((and (= (vector-ref map 0) -1) (not (= (vector-ref map 2) -1)))
	(vector-set! map 0 1)
	(vector-set! map 1 1))
       ((and (= (vector-ref map 2) -1) (not (= (vector-ref map 0) -1)))
	(vector-set! map 2 1)
	(vector-set! map 3 1)))
      map)))

(define (need-dictref)
  (cond
   ((xmltrans:attr "TABULAR" "DICTREF") =>
    (lambda (x)
      (let ((map (parse-dictref x)))
	(and
	 (not (= (map-start-col map) -1))
	 (or
	  (= (map-start-col map) 0)
	  (<= (map-start-col map) (eff-col-index)))
	 (or
	  (= (map-end-col map) 0)
	  (>= (map-end-col map) (eff-col-index)))
	 (not (= (map-start-row map) -1))
	 (or
	  (= (map-start-col map) 0)
	  (<= (map-start-row map) (eff-row-index)))
	 (or
	  (= (map-end-row map) 0)
	  (>= (map-end-row map) (eff-row-index)))))))
   (else
    #f)))

(define (get-col-align ind)
  (let ((cl (xmltrans:attr "TABULAR" "COLALIGN")))
    (if cl
	(let ((s (string-split cl #\:)))
	  (cond
	   ((null? (cdr s))
	    cl)
	   ((<= (length s) ind)
	    (if (string=? (car (last-pair s)) "*")
		(list-ref s (- (length s) 2))
		#f))
	   ((string=? (list-ref s ind) "*")
	    (list-ref s (- ind 1)))
	   ((string-null? (list-ref s ind))
	    #f)
	   (else
	    (list-ref s ind))))
	#f)))

(xmltrans:end-tag
 "ITEM"
 (tag attr text)
 (cond
  ((or (xmltrans:parent? "ENUMERATE") (xmltrans:parent? "ITEMIZE"))
   (list "<LI>" text "</LI>"))
  ((xmltrans:parent? "DEFLIST")
   (list "<DD>" text "</DD>"))
  ((xmltrans:parent? "ROW")
   (letrec ((sl (list))
	    (put (defn-put put sl)))
     (let ((eltname (if (and
			 (= row-index 0)
			 (xmltrans:attr "TABULAR" "COLHEADING"))
			"TH"
			"TD")))

       (put "<" eltname)
       (let ((span (xmltrans:attr "ITEM" "SPAN")))
	 (cond
	  (span
	   (let ((n (if (string=? span "0")
			max-items
			span)))
	     (put " COLSPAN=\"" span "\" ALIGN=\"CENTER\"")))
	  (else
	   (cond
	    ((string=? eltname "TD")
	     (let ((align (get-col-align item-index)))
	       (if align
		   (put " ALIGN=\"" align "\"")))
	     (let ((rowspan (xmltrans:attr "ITEM" "ROWSPAN")))
	       (cond
		((and (= item-index 0) rowspan)
		 (set! rowspan-end (+ row-index
				      (string->number rowspan)))
		 (put " ROWSPAN=\"" rowspan "\""))))))))

	 (let ((rowheading (and (= item-index 0)
				(xmltrans:attr "TABULAR" "ROWHEADING"))))
	   (if rowheading
	       (put " CLASS=\"" rowheading "\""))

	   (put ">")

	   (cond
	    (rowheading
	     (put text))
	    (else
	     (let ((x (need-dictref)))
	       (if x
		   (put (dict-reference (restore-word text)
					(item-text text)
					x))
		   (put (item-text text))))))
	   (put "</" eltname ">")
	   (set! item-index (1+ item-index))
	   (if (< max-items item-index)
	       (set! max-items item-index)))))

     (reverse sl)))
  ((xmltrans:parent? "GROUP")
   (let ((fn (xmltrans:attr "LINGUA" text)))
     (cond
      ((not fn)
       (xmltrans:parse-error #f "Function " text " not defined"))
      ((not (procedure? fn))
       (xmltrans:parse-error #f text " is not a function"))
      (else
       (menu-append #t fn ""))))
   #f)
  ((xmltrans:parent? "PAGE")
   (list "<H2>" text "</H2>"))
  (else
   (list "<DD>" text "</DD>"))))

(xmltrans:end-tag
 "FLECT"
 (tag attr text)
 (cond
  ((assoc "SPLIT" attr) =>
   (lambda (delim)
     (let ((sl (list)))
       (letrec ((put (defn-put put sl)))
	 (put "<B>")
	 (let ((tok-list (string-split text (string-ref (cdr delim) 0))))
	   (put "-" (car tok-list))
	   (for-each
	    (lambda (s)
	      (put "&nbsp;,&nbsp;-" s))
	    (cdr tok-list)))
	 (put "</B>")
	 (reverse sl)))))
  (else
   (list "<B>-" text "</B>"))))

(xmltrans:end-tag
 "ALPH"
 (tag attr text)
 (list "<span class=\"literal\">" text "</span>"))

(xmltrans:end-tag
 "IPA"
 (tag attr text)
 (list "<span class=\"ipa\">" text "</span>"))

(xmltrans:end-tag
 "DICTREF"
 (tag attr text)
 (dict-reference text text))

;;; ENUMERATE
(xmltrans:end-tag
 "ENUMERATE"
 (tag attr text)
 (list
  (cond
   ((xmltrans:parent? "PARA")
    (xmltrans:set-attr "PARA" "__CLOSED__" 1)
    "\n</P>\n")
   (else
    ""))
  "<OL>\n" text "</OL>\n"))

(xmltrans:end-tag
 "ITEMIZE"
 (tag attr text)
 (list
  (cond
   ((xmltrans:parent? "PARA")
    (xmltrans:set-attr "PARA" "__CLOSED__" 1)
    "\n</P>\n")
   (else
    ""))
  "<UL>\n" text "</UL>\n"))


(xmltrans:end-tag
 "DEFLIST"
 (tag attr text)
 (list
  (cond
   ((xmltrans:parent? "PARA")
    (xmltrans:set-attr "PARA" "__CLOSED__" 1)
    "\n</P>\n")
   (else
    ""))
  "<DL>\n" text "</DL>\n"))

(xmltrans:start-tag
 "TABULAR"
 (tag attr)
 (letrec ((sl (list))
	  (put (defn-put put sl)))
   (if (xmltrans:parent? "PARA")
       (begin
	 (xmltrans:set-attr "PARA" "__CLOSED__" 1)
	 (put "\n</P>\n")))
   
   (let ((sp (xmltrans:attr #f "SPACE")))
     (if sp
	 (do ((i 0 (1+ i)))
	     ((= i (string->number sp)))
	   (put "<BR>"))))

   (put (cond
	 ((xmltrans:attr #f "ALIGN") =>
	  (lambda (align)
	    (if (not (string-ci=? align "NONE"))
		(alignment align))))
	 (else
	  (alignment "CENTER"))))

   (put "<TABLE CELLPADDING=\"10%\" ")

   (let ((type (or (xmltrans:attr #f "TYPE") "")))
     (cond
      ((string=? type "EXAMPLE")
       (xmltrans:set-attr #f "NOFRAME" 1)))

     (if (not (xmltrans:attr #f "NOFRAME"))
	 (put " frame=\"box\""))
     (if (not (xmltrans:attr #f "BESTFIT"))
	 (put " WIDTH=\"95%\""))

     (cond
      ((xmltrans:attr #f "CLASS") =>
       (lambda (class)
	 (put " class=\"" class "\""))))

     (put ">\n")

     (if (string=? type "EXAMPLE")
	 (put "<COL SPAN=2 WIDTH=\"50%\">\n"))

     (set! row-index 0)
     (set! max-items 0)
     (set! separator-row 0)

     (reverse sl))))

(define table-number 0) ;; Ordinal number of the current table

(xmltrans:end-tag
 "TABULAR"
 (tag attr text)
 (set! table-number (1+ table-number))
 (let ((sl (list)))
   (letrec ((put (defn-put put sl)))
     (put text)
     (put "</TABLE>\n")

     (let ((align (xmltrans:attr #f "ALIGN")))
       (if (or (not align) (not (string-ci=? align "NONE")))
	   (put (alignment ""))))

     (let ((sp (xmltrans:attr #f "AFTERSPACE")))
       (if sp
	   (do ((i 0 (1+ i)))
	       ((= i (string->number sp)))
	     (put "<BR>"))))

     (let ((x (xmltrans:attr #f "__COLSPAN__")))
       (if x
	   (let ((max max-items))
	     (letrec ((proc (lambda ()
			      (display max))))
	       (gamlet:set-handler x proc)))))

     (let ((fl (xmltrans:attr #f "FOOTNOTES")))
       (map-footnotes fl put))

     (reverse sl))))

;;; ROW

(xmltrans:start-tag
 "ROW"
 (tag attr)
 (let ((sl (list)))
   (letrec ((put (defn-put put sl)))
     (set! item-index 0)
     (cond
      ((and (= row-index 0) (xmltrans:attr "TABULAR" "COLHEADING")) =>
       (lambda (class)
	 (put "<THEAD class=\"" class "\">" "<TR>\n")))
      (else
       (put "<TR")
       (if (xmltrans:attr "TABULAR" "ALTERNATE")
	   (put " class=\"" (if (= (modulo (- row-index separator-row) 2) 0)
				"even"
				"odd")
		"\""))
       (put ">\n")))
     (if (< row-index rowspan-end)
	 (set! item-index (1+ item-index)))
     (reverse sl))))

(xmltrans:end-tag
 "ROW"
 (tag attr text)
 (let ((retval (list text "</TR>")))
   (if (and (= row-index 0)
	    (xmltrans:attr "TABULAR" "COLHEADING"))
       (set! retval (append retval (list "</THEAD>"))))

   (set! row-index (1+ row-index))
   (if (= row-index rowspan-end)
       (set! rowspan-end 0))
   
   (append retval (list "\n"))))
   

(define separator-count 0) ;; Number of separators for which a GAMLET macro
                           ;; has been created.

;;; SEPARATOR   
(xmltrans:start-tag
 "SEPARATOR"
 (tag attr)
 (let ((max (cond
	     ((xmltrans:attr #f "SPAN") =>
	      (lambda (x)
		x))
	     ((= row-index 0)
	      ;; This is the first row. We still do not know how many columns
	      ;; will it have.
	      (let ((x (xmltrans:attr "TABULAR" "__COLSPAN__")))
		(if x
		    x
		    (let ((name (string-append
				 "@@COLSPAN:"
				 (number->string separator-count))))
		      (set! separator-count (1+ separator-count))
		      (xmltrans:set-attr "TABULAR" "__COLSPAN__" name)
		      (let ((delim (gamlet:get-pp-delimiter)))
			(format #f "~A~A~A" (car delim) name (cdr delim)))))))
	     (else
	      (number->string max-items)))))
   (list "\n<TR><TD COLSPAN=\"" max "\" class=\"separator\">")))

(xmltrans:end-tag
 "SEPARATOR"
 (tag attr text)
 (set! separator-row (if (or (xmltrans:attr #f "INITCOUNT")
			     (xmltrans:attr "TABULAR" "INITCOUNT"))
			 0
			 row-index))
 (list text "</td></tr>\n"))
     

;;;

(define (class-begin attr class-name)
  (letrec ((sl (list))
	   (put (defn-put put sl)))
    (let ((tag (cond
		((xmltrans:parent? "ROW")
		 "TD")
		((xmltrans:parent? "TABULAR")
		 (put "<TR class=\"" class-name "\""))
		((xmltrans:attr attr "TYPE") =>
		 (lambda (type)
		   (if (string=? type "INLINE")
		       "SPAN"
		       "DIV")))
		((xmltrans:attr attr "ALIGN") =>
		 (lambda (x)
		   (string-append "DIV ALIGN=\"" x "\"")))
		(else
		 "SPAN"))))
      (put "<" tag " class=\"" class-name "\">")
      (reverse sl))))

(define (class-end attr)
  (letrec ((sl (list))
	   (put (defn-put put sl)))
    (let ((tag (cond
		((xmltrans:parent? "ROW")
		 "TD")
		((xmltrans:parent? "TABULAR")
		 "TD")
		((xmltrans:attr attr "TYPE") =>
		 (lambda (type)
		   (if (string=? type "INLINE")
		       "SPAN"
		       "DIV")))
		((xmltrans:attr attr "ALIGN")
		 "DIV")
		(else
		 "SPAN"))))
      (put "</" tag ">")
      (if (xmltrans:parent? "TABULAR")
	  (put "</TR>"))
      (reverse sl))))

(xmltrans:end-tag
 "EXAMPLE"
 (tag attr text)
 (let ((type (xmltrans:attr attr "TYPE")))
   (if (and type (not (string=? type "INLINE")))
       (list
	"<div align=\"center\">"
	(class-begin #f "example")
	text
	(class-end #f)
	"</div>")
       ; else
       (list
	(class-begin #f "example")
	text
	(class-end #f)))))

(xmltrans:end-tag
 "TRANS"
 (tag attr text)
 (list
  (class-begin #f "translation")
  text
  (class-end #f)))

(xmltrans:end-tag
 "SAMP"
 (tag attr text)
 (list
  "<SPAN class=\"sample\">"
  text
  "</SPAN>"))

(xmltrans:end-tag
 "NOTREADY"
 (tag attr text)
 (list "
 <table class=\"notready\">
  <tr>
   <td><img border=\"0\" src=\"graphics/construction.jpg\" alt=\"[ Construction Bar ]\" width=\"300\" height=\"28\"></img></td>
<td class=\"notready\">"
     text
     "
   </td>
   <td><img border=\"0\" src=\"graphics/construction.jpg\" alt=\"[ Construction Bar ]\" width=\"300\" height=\"28\"></img></td>
   </tr>
</table>"))


;;;
(define (alignment a)
  (if (string-null? a)
      "</DIV>\n"
      (string-append "<DIV ALIGN=\"" a "\">")))


;;; Postprocessor handlers

(define (toc-pp)
  (out-menu (lingua:attr "LINGUA" "MENU") 0))

(gamlet:set-handler "TOC" toc-pp)

(define (sitemap-pp)
  (print-menu-map (lingua:attr "LINGUA" "MENU")))

(gamlet:set-handler "SITEMAP" sitemap-pp)

(define (info-pp)
  ;; Empty function. The user is supposed to override it if he needs to
  #f)
  
(gamlet:set-handler "INFO" info-pp)      

(define (bottom-navbar-pp)
  (display "<div class=\"bottom-navbar\">")
  (let ((index (gamlet:output-file-index)))
    (display "<table class=\"bottom-navbar\" width=\"100%\">
<tr>
 <td align=\"left\" width=\"33%\">")
    (if (> index 0)
	(format #t "<a href=\"~A\"><img src=\"graphics/left.png\" alt=\"[ ~A ]\" border=\"0\">~A</a>"
		(reference-name
		 (file-name->reference (gamlet:output-file-name (- index 1))))
		(lingua:gettext "Previous Page")
		(find-file-in-menu
		 (file-name->reference (gamlet:output-file-name (- index 1))))))

    (display "</td><td align=\"center\" width=\"33%\">")
    (format #t "<a href=\"~A\#top\"><img src=\"graphics/up.png\" alt=\"[ ~A ]\" border=\"0\"><br>~A</a>"
	    (reference-name
	     (file-name->reference (gamlet:current-output-file-name)))
	    (lingua:gettext "Top of the Page")
	    (lingua:gettext "Top of the Page"))

    (display "</td><td align=\"right\" width=\"33%\">")
    
    (if (< index (- (gamlet:output-file-count) 1))
	(format #t "<a href=\"~A\">~A<img src=\"graphics/right.png\" alt=\"[ ~A ]\" border=\"0\"></a>"
		(reference-name
		 (file-name->reference (gamlet:output-file-name (+ index 1))))
		(find-file-in-menu
		 (file-name->reference (gamlet:output-file-name (+ index 1))))
		(lingua:gettext "Next Page")))

    (display "</td>
</table>
</div>")))

(gamlet:set-handler "BOTTOM-NAVBAR" bottom-navbar-pp)      


;;;; Command line parser
(define (cons? p)
  (and (pair? p) (not (list? p))))

(define (usage)
  (display "usage: lingua.scm OPTIONS\n")
  (display "OPTIONS are:\n")
  (display "  --base-href URL\n")
  (display "  --prefix DIR\n")
  (display "  --config-file FILE\n")
  (display "  --output-dir DIR\n")
  (display "  --tmp-dir DIR\n")
  (display "  --debug NUMBER\n")
  0)

(define (parse-command-line arglist)
  (let ((grammar `((base-href (single-char #\b)
			      (value #t))
		   (config-file (single-char #\c)
				(value #t))
		   (prefix    (single-char #\p)
			      (value #t))
		   (output-dir (single-char #\o)
			       (value #t))
		   (tmp-dir   (single-char #\t)
			      (value #t))
		   (debug (single-char #\d)
			  (value #t))
		   (help (single-char #\h))))
	(input-files '()))
    (for-each
     (lambda (x)
       (cond
	((cons? x)
	 (case (car x)
	   ((debug)
	    (xmltrans:set-debug-level (string->number (cdr x))))
	   ((config-file)
	    (set! config-must-exist #t)
	    (set! config-file-name (cdr x)))
	   ((help)
	    (exit (usage)))
	   ((prefix)
	    (set! prefix-option (cdr x)))
	   ((base-href)
	    (set! base-href-option (cdr x)))
	   ((output-dir)
	    (set! output-dir-option (cdr x)))
	   ((tmp-dir)
	    (set! tmp-dir-option (cdr x)))))
	(else
	 (set! input-files (cdr x)))))
     (getopt-long arglist grammar))
    input-files))

(define (begin-processing)
  ;; First, see if we have a configuration file
  (cond
   ((file-exists? config-file-name)
    (if (not (xmltrans:parse-file config-file-name))
	(exit 1)))
   (config-must-exist
    (display (string-append
	      "Configuration file " config-file-name " not found\n")
	     (current-error-port))
    (exit 1)))

  (cond
   ((or output-dir-option
	(and (lingua:tag-exists? "DIRECTORY")
	     (lingua:attr "DIRECTORY" "OUT"))) =>
	     (lambda (x)
	       (gamlet:set-output-dir x))))

  (cond
   ((or tmp-dir-option
	(and (lingua:tag-exists? "DIRECTORY")
	     (lingua:attr "DIRECTORY" "TMP"))) =>
	     (lambda (x)
	       (gamlet:set-working-dir x)))))

(define (end-processing)
  #t)

;;;; End of lingua.scm

   
